using System;
using System.Reflection;
using System.Linq;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;


namespace Bitsbolts.Components {
  // This class is effectively internal, but Monocle [Tracked] needs public
  [Tracked]
  public class AwakePriority : Component {
    [OnLoad]
    internal static void Load() {
      IL.Monocle.EntityList.UpdateLists += OnEntityListUpdateLists;
    }

    [OnUnload]
    internal static void Unload() {
      IL.Monocle.EntityList.UpdateLists -= OnEntityListUpdateLists;
    }


    private static bool Constructed = false;

    internal int Priority;
    internal AwakePriority(int priority) : base(false, false) {
      this.Priority = priority;
      AwakePriority.Constructed = true;
    }

    private static void OnEntityListUpdateLists(ILContext il) {
      var cursor = new ILCursor(il);

      // for whatever reason callvirt doesn't match here so fuck it use Ldfld
      if (!cursor.TryGotoNext(MoveType.Before, instr => instr.MatchLdfld<EntityList>("adding"))) {
        Logger.Log(LogLevel.Error, "bitsbolts.Components.AwakePriority", $"IL@{cursor.Index}: Hook failed to find ldfld EntityList.adding.Clear()");
        return;
      }

      Logger.Log("bitsbolts.Components.AwakePriority", $"IL@{cursor.Index}: Adding delegate callback in EntityList after toAwake.AddRange!");
      cursor.Emit(OpCodes.Ldarg_0);
      cursor.EmitDelegate<Action<EntityList>>(
        (EntityList entityList) => {
          // so uhh fun fact: Everest's crash screen can trigger in the middle
          // of Initialize, which means that Engine.Scene can literally fucking
          // be null??
          if (!(Engine.Scene != null && Engine.Scene.Tracker != null && Constructed)) {
            // if the tracker isn't ready, then this will component won't be effective anyways
            return;
          }

          DynamicData.For(entityList).TryGet<List<Entity>>("toAwake", out List<Entity> toAwake);

          // C#'s stable sort is slow for our usecase; use counting sort instead
          var sortedAwake = Collections.CountingSort(
            toAwake,
            (Entity entity) => entity.Get<AwakePriority>()?.Priority ?? 0
          ).ToList();

          toAwake.Clear();
          toAwake.AddRange(sortedAwake);
        }
      );
    }
  }
}
