using System;
using System.Collections;
using System.Reflection;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.Utils;
using Celeste;
using Celeste.Editor;
using Celeste.Mod;
using Celeste.Mod.Entities;


namespace Bitsbolts.Entities {
  [MapDataTracked("Bitsbolts/DebugCoreMessage")]
  [CustomEntity("Bitsbolts/DebugCoreMessage")]
  public class DebugCoreMessage : Entity {
    [OnLoad]
    public static void Load() {
      On.Celeste.Editor.MapEditor.Render += OnMapEditorRender;
    }

    [OnUnload]
    public static void Unload() {
      On.Celeste.Editor.MapEditor.Render -= OnMapEditorRender;
    }


    static float FadeDistance = 512f;

    internal FancyText.Text Text;
    private String dialog;

    private float alpha;

    public DebugCoreMessage(EntityData data, Vector2 offset) : base(data.Position + offset) {
      base.Collidable = false;
      base.Collider = new Hitbox(data.Width, data.Height, 0f, 0f);

      this.Tag |= TagsExt.SubHUD | Tags.PauseUpdate | Tags.FrozenUpdate | Tags.TransitionUpdate;
      this.dialog = data.Attr("dialog", "{no dialog set dumbass}").DialogNlOrKey();
    }

    public override void Added(Scene scene) {
      base.Added(scene);

      Level level = this.Scene as Level;
      var width = Vector2.Distance(
        level.WorldToScreen(this.TopRight),
        level.WorldToScreen(this.TopLeft)
      );

      this.Text = FancyText.Parse(
        text: this.dialog,
        maxLineWidth: (int) width,
        linesPerPage: 40
      );

      if (!(scene as Level).Session.Area.SID.StartsWith("bitsbolts/")) {
        throw new Exception("Dumbass this entity is internal ONLY!");
      }
    }

    public override void Update() {
      base.Update();

      // Player player = Scene.Tracker.GetEntity<Player>();
      var camera = (this.Scene as Level).Camera;
      if (camera != null) {
        this.alpha = Ease.CubeInOut(Calc.ClampedMap(Math.Abs(this.X - camera.X), 0f, FadeDistance, 1f, 0f));
      }
    }

    public override void Render() {
      Level level = this.Scene as Level;

      Vector2 renderPosition = level.WorldToScreen(this.TopCenter);

      this.Text.DrawJustifyPerLine(
        position: renderPosition,
        justify: new Vector2(0.5f, 0f),
        scale: Vector2.One,
        alpha: this.alpha
      );
    }

    public static void OnMapEditorRender(On.Celeste.Editor.MapEditor.orig_Render orig, MapEditor editor) {
      orig(editor);

      var tracker = DynamicData.For(editor).Get<MapData>("mapData").GetTracker();
      var camera = (Camera) typeof(MapEditor).GetField("Camera", BindingFlags.NonPublic | BindingFlags.Static).GetValue(null);

      Draw.SpriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend, SamplerState.LinearClamp, DepthStencilState.None, RasterizerState.CullNone, null, Engine.ScreenMatrix);


      foreach (var data in tracker.GetEntityDatas("Bitsbolts/DebugCoreMessage")) {
        var position = data.Level.Position + data.Position + (new Vector2(data.Width, data.Height) * .5f);
        var text = ActiveFont.FontSize.AutoNewline(
          data.Attr("dialog").FormatPlain(),
          (int) data.Width * 8
        );
        ActiveFont.DrawOutline(
          text,
          camera.CameraToScreen(position / 8f),
          new Vector2(.5f, .5f),
          Vector2.One * (camera.Zoom * .025f),
          Color.White,
          2f,
          Color.Black
        );
      }

      Draw.SpriteBatch.End();
    }
  }
}
