using System;
using System.Reflection;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using Bitsbolts.Hooks;


namespace Bitsbolts.Entities {
  [CustomEntity("Bitsbolts/EnableTriggersInReflectionFallController")]
  [Tracked]
  public class EnableTriggersInReflectionFallController : Entity {
    private static IDetour ILPlayerUpdateHook;

    [OnLoad]
    internal static void Load() {
      ILPlayerUpdateHook = new ILHook(
        typeof(Player).GetMethod("orig_Update", BindingFlags.Public | BindingFlags.Instance),
        ILPlayerUpdate
      );
    }

    [OnUnload]
    internal static void Unload() {
      ILPlayerUpdateHook.Dispose();
    }

    public EnableTriggersInReflectionFallController(EntityData data, Vector2 offset) : base() {
      this.Active = false;
      this.Collidable = false;
      this.Visible = false;
    }


    private static void ILPlayerUpdate(ILContext il) {
      var cursor = new ILCursor(il);

      if (!cursor.TryGotoNext(MoveType.After, instr => instr.MatchCallOrCallvirt<Player>("UpdateSprite"))) {
        Logger.Log(LogLevel.Error, "Bitsbolts.Entities.EnableTriggersInReflectionFallController", $"IL@{cursor.Index}: Hook failed to jump to callvirt player.UpdateSprite() in Player.Update.");
        return;
      }

      if (!cursor.TryGotoNext(
        MoveType.Before,
        instr => instr.MatchLdfld<Player>("StateMachine"),
        instr => instr.MatchCallvirt<StateMachine>("get_State"),
        instr => instr.MatchLdcI4(Player.StReflectionFall)
      )) {
        Logger.Log(LogLevel.Error, "Bitsbolts.Entities.EnableTriggersInReflectionFallController", $"IL@{cursor.Index}: Hook failed to find stfld this.StateMachine.State != 18 in Player.Update.");
        return;
      }

      cursor.GotoNext();
      cursor.GotoNext();

      Logger.Log("Bitsbolts.Entities.EnableTriggersInReflectionFallController", $"IL@{cursor.Index}: Adding delegate in Player.Update to mod this.StateMachine.State != 18");

      cursor.Emit(OpCodes.Ldarg_0);
      cursor.EmitDelegate<Func<int, Player, int>>(
        (state, player) => {
          return (
            state == Player.StReflectionFall
            && player.Scene.Tracker.CountEntities<EnableTriggersInReflectionFallController>() > 0
          ) ? state + 1 : state;
        }
      );
    }
  }
}
