using System;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using Bitsbolts.Hooks;


namespace Bitsbolts.Entities {
  [MapDataTracked("Bitsbolts/GlobalRoomLoaderController")]
  public class GlobalRoomLoaderController {
    [OnLoad]
    internal static void Load() {
      Everest.Events.Level.OnLoadEntity += Load;
    }

    [OnUnload]
    internal static void Unload() {
      Everest.Events.Level.OnLoadEntity -= Load;
    }

    public static bool Load(Level level, LevelData levelData, Vector2 offset, EntityData data) {
      if (data.Name != "Bitsbolts/GlobalRoomLoaderController") { return false; }

      // We can't use level directly, because we might belong to a room other
      // than level, but are regardlessly being loaded in level. This happens
      // when using GlobalRoomLoader (which is obviously relevant)
      var eid = new EntityID(data.Level.Name, data.ID);

      if (!eid.Level.StartsWith(GlobalRoomLoader.GLOBAL_ROOM_PREFIX)) {
        throw new Exception($"GlobalRoomLoaderControllers must be placed in a room which has a name that starts with {GlobalRoomLoader.GLOBAL_ROOM_PREFIX} but the controller with id {eid} is not placed properly!");
      }

      return true;
    }

    internal readonly Dictionary<string, bool> Flags;
    internal readonly Dictionary<string, LevelData> Rooms;
    internal readonly bool DuringTransition = true;

    internal readonly static GlobalRoomLoaderController DEFAULT_INSTANCE = new GlobalRoomLoaderController();

    internal GlobalRoomLoaderController() {
      this.Flags = new Dictionary<string, bool>();
      this.Rooms = null;
      this.DuringTransition = true;
    }


    internal GlobalRoomLoaderController(EntityData data, MapData map) {
      this.Flags = data.Attr("flags", "").ParseFlagstrs(allowInversions: true);

      var rooms = data.Attr("rooms", "");
      this.Rooms = rooms != "" ? rooms.ParseLevelsList(map) : null;

      this.DuringTransition = !data.Bool("notDuringTransition", false);
    }
  }
}
