using System;
using System.Linq;

using Microsoft.Xna.Framework;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;


namespace Bitsbolts.Entities {
  [CustomEntity("Bitsbolts/GrablessBerryBlock")]
  [TrackedAs(typeof(GoldenBlock))]
  public class GrablessBerryBlock : GoldenBlock {
    [OnLoad]
    public static void Load() {
      IL.Celeste.GoldenBlock.ctor_Vector2_float_float += ILGoldenBlockCtor;
      IL.Celeste.GoldenBlock.Update += ILGoldenBlockUpdate;
    }

    [OnUnload]
    public static void Unload() {
      IL.Celeste.GoldenBlock.ctor_Vector2_float_float -= ILGoldenBlockCtor;
    }

    public bool Floaty;
    public GrablessBerryBlock(EntityData data, Vector2 offset) : base(data, offset) {
      this.Floaty = data.Bool("floaty", true);
    }

    public override void Awake(Scene scene) {
      // mm yes this base.Awake is made of Scene
      if (this.Components != null) {
        foreach (Component component in this.Components) {
          component.EntityAwake();
        }
      }

      // idk why GoldenBlock waits until player > this.X - 80; stop that shit
      base.Visible = false;
      base.Collidable = false;
      base.DisableStaticMovers();
      DynamicData.For(this as GoldenBlock).Set("renderLerp", 1f);
    }

    public override void Update() {
      var enabled = base.Scene.Tracker.GetEntities<GrablessBerry>().Any(
        (Entity entity) =>
          (entity as GrablessBerry).Follower.Leader != null
          || (entity as GrablessBerry).TagCheck(Tags.Global | Tags.Persistent)
      );

      if (enabled != base.Collidable) {
        DynamicData.For(this as GoldenBlock).Set("renderLerp", 1f);
        if (enabled) {
          base.EnableStaticMovers();
        } else {
          base.DisableStaticMovers();
        }
      }
      base.Visible = enabled;
      base.Collidable = enabled;

      if (base.Visible) {
        base.Update();
      }
    }

    private static void ILGoldenBlockCtor(ILContext il) {
      ILCursor cursor = new ILCursor(il);

      if (!cursor.TryGotoNext(
        MoveType.After,
        instr => instr.MatchLdstr("collectables/goldberry/idle00")
      )) {
        Logger.Log(LogLevel.Error, "bitsbolts.GrablessBerryBlock", $"IL@{cursor.Index}: Hook failed to find \"collectables/goldberry/idle00\" in GoldenBlock");
        return;
      }
      Logger.Log("bitsbolts.GrablessBerryBlock", $"IL@{cursor.Index}: Adding delegate callback in GoldenBlock to mod \"collectables/goldberry/idle00\"");

      cursor.Emit(OpCodes.Ldarg_0);
      cursor.EmitDelegate<Func<string, GoldenBlock, string>>(
        (string orig, GoldenBlock self) =>
          self is GrablessBerryBlock ? "bitsbolts/grabless_berry/idle00" : orig
      );


      if (!cursor.TryGotoNext(
        MoveType.After,
        instr => instr.MatchLdstr("objects/goldblock")
      )) {
        Logger.Log(LogLevel.Error, "bitsbolts.GrablessBerryBlock", $"IL@{cursor.Index}: Hook failed to find \"objects/goldblock\" in GoldenBlock");
        return;
      }
      Logger.Log("bitsbolts.GrablessBerryBlock", $"IL@{cursor.Index}: Adding delegate callback in GoldenBlock to mod \"objects/goldblock\"");

      cursor.Emit(OpCodes.Ldarg_0);
      cursor.EmitDelegate<Func<string, GoldenBlock, string>>(
        (string orig, GoldenBlock self) =>
          self is GrablessBerryBlock ? "bitsbolts/grabless_berry/block" : orig
      );
    }

    private static void ILGoldenBlockUpdate(ILContext il) {
      ILCursor cursor = new ILCursor(il);

      if (!cursor.TryGotoNext(
        MoveType.After,
        instr => instr.MatchStfld<GoldenBlock>("sinkTimer")
      )) {
        Logger.Log(LogLevel.Error, "bitsbolts.GrablessBerryBlock", $"IL@{cursor.Index}: Hook failed to find stfld this.sinkTimer in GoldenBlock");
        return;
      }
      Logger.Log("bitsbolts.GrablessBerryBlock", $"IL@{cursor.Index}: Adding delegate callback in GoldenBlock after stfld this.sinkTimer");

      cursor.Emit(OpCodes.Ldarg_0);
      cursor.EmitDelegate<Action<GoldenBlock>>((GoldenBlock self) => {
        if (self is GrablessBerryBlock) {
          DynamicData.For(self).Set("sinkTimer", 0f);
        }
      });
    }
  }
}
