using System;
using System.Linq;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using Bitsbolts.Components;
using Bitsbolts.Hooks;


// todo: maybe an entity for screen wrapping? (needs unlocking camera first!)
// todo: if it should move backdrops, make a non-teleporting camera entity
// todo: do something for stylegrounds
//   (maybe if LoopX then Backdrop.Position.X += Delta.X, and viseversa for Y)
//   (idk what to do with non-looping stylegrounds)

namespace Bitsbolts.Entities {
  [Tracked]
  [CustomEntity("Bitsbolts/WorldPortal")]
  public class WorldPortal : Entity {
    internal Components.WorldCamera Camera;

    public WorldPortal(EntityData data, Vector2 offset) : base(data.Position + offset) {
      base.Visible = false;
      base.Collider = new Hitbox(data.Width, data.Height);

      base.Add(this.Camera = new Components.WorldCamera(
        position: Vector2.Zero,
        target: data.NodesOffset(offset)[0],
        width: data.Width,
        height: data.Height
      ));

      // fixme: we don't use Visible nor Collidable!
      var flags = data.Attr("flags").ParseFlagstrs();
      if (flags.Count > 0) {
        base.Add(new FlagstrListener {
          Flags = flags, TriggerOnAwake = true, OnAll = (enabled) => {
            this.Camera.Active = enabled;
            this.Collidable = enabled;
          }
        });
      }
    }


    public override void Update() {
      base.Update();
      Level level = this.Scene as Level;

      if (!this.Collidable) { return; }
      // foreach (Entity collided in this.CollideAllExcept(Tags.HUD)) {
      foreach (Entity collided in this.CollideAll<Actor>()) {
        if (collided is Player player) {
          level.TeleportSeamlessly(player, this.Camera.Delta);
        } else {
          collided.Position += this.Camera.Delta;
        }
      }

      foreach (Entity collided in this.CollideAll<Solid>()) {
        if (!(collided is SolidTiles)) {
          collided.Position += this.Camera.Delta;
        }
      }
    }
  }
}
