using System;
using System.Linq;
using System.Reflection;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;


namespace Bitsbolts {
  internal static class Events {
    private static IDetour OnCelesteMapDataLoadHook;

    [OnLoad]
    private static void Load() {
      On.Celeste.Level.LoadLevel += OnLevelLoadLevel;

      OnCelesteMapDataLoadHook = new Hook(
        typeof(MapData).GetMethod("Load", BindingFlags.NonPublic | BindingFlags.Instance),
        typeof(Events).GetMethod("OnCelesteMapDataLoad", BindingFlags.NonPublic | BindingFlags.Static)
      );
    }

    [OnUnload]
    private static void Unload() {
      On.Celeste.Level.LoadLevel -= OnLevelLoadLevel;
      OnCelesteMapDataLoadHook.Dispose();
    }



    internal delegate void OnMapDataLoadDelegate(MapData map);
    internal static event OnMapDataLoadDelegate OnMapDataLoad;
    private static void MapDataLoad(MapData mapData) {
      MapDataTracker.OnMapDataLoad(mapData);
      OnMapDataLoad(mapData);
    }


    private static void OnCelesteMapDataLoad(On.Celeste.MapData.orig_Load orig, MapData mapData) {
      orig(mapData);
      DynamicData.For(mapData).Set("_bitsbolts_event_mapdataloaded", (bool?) true);
      MapDataLoad(mapData);
    }


    // This is a backup specifically for randomizer when MapData.Load doesn't get
    // called. We tag mapdatas as we encounter them this way, and we call load
    // if we find untagged ones. Of course this doesn't handle Reload, but
    // randomizer never calls Reload anyways so whatever...
    private static void OnLevelLoadLevel(
      On.Celeste.Level.orig_LoadLevel orig, Level level, Player.IntroTypes intro, bool isFromLoader
    ) {
      var mapData = level.Session.MapData;
      var dd = DynamicData.For(mapData);

      if ((bool) (dd.Get<bool?>("_bitsbolts_event_mapdataloaded") ?? false) == false) {
        dd.Set("_bitsbolts_event_mapdataloaded", (bool?) true);
        MapDataLoad(mapData);
      }

      orig(level, intro, isFromLoader);
    }
  }
}
